/*
 * outpost
 *   <mycroft@mastersec.org>
 */

/* software dedicated to all my friends of #epita.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/param.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <pcap.h>
#include <getopt.h>
#include <dlfcn.h>
#include <errno.h>
#include <syslog.h>

#include "outpost.h"
#include "common.h"
#include "modules.h"
#include "config.h"
#include "interfaces.h"

#define VERSION "0.0.1-rc3"

#ifndef NOFILE
#define NOFILE 256
#endif

static struct option long_options[] =
{
  { "conftest", 0, 0, 'c' },
  { "daemon", 0, 0, 'd' },
  { "help", 0, 0, 'h' },
  { "ignoreconf", 0, 0, 'I' },
  { "interface", 0, 0, 'i' },
  { "module",  1, 0, 'm' },
  { "pidfile", 1, 0, 'p' },
  { "read", 1, 0, 'r' },
  { "snaplen", 1, 0, 's' },
  { "testbpf", 0, 0, 't' },
  { "verbose", 0, 0, 'v' }, /* XXX */
  { "write", 1, 0, 'w' },
  { "testrealbpf", 0, 0, 'T' },
  { "version", 0, 0, 'V' },
  { 0, 0, 0, 0 }
};

pcap_t *desc = NULL;
pcap_dumper_t *logdesc = NULL;

int
usage (void)
{
  _print (STYLE_COMMON, "outpost - by cns at minithins.net\n");
  _print (STYLE_COMMON, "outpost <options> [bpf rule]\n");
  _print (STYLE_COMMON, "options:\n");
  _print (STYLE_COMMON, "  -c  --conftest\n");
  _print (STYLE_COMMON, "  -d  --daemon\n");
  _print (STYLE_COMMON, "  -h  --help\n");
  _print (STYLE_COMMON, "  -i  --interface [interface]\n");
  _print (STYLE_COMMON, "  -m  --module [module]\n");
  _print (STYLE_COMMON, "  -p  --pidfile [file]\n");
  _print (STYLE_COMMON, "  -r  --read [pcap file]\n");
  _print (STYLE_COMMON, "  -s  --snaplen [len]\n");
  _print (STYLE_COMMON, "  -t  --testbpf\n");
  _print (STYLE_COMMON, "  -w  --write [pcap file]\n");
  _print (STYLE_COMMON, "  -I  --ignoreconf\n");
  _print (STYLE_COMMON, "  -T  --testrealbpf\n");
  _print (STYLE_COMMON, "  -V  --version\n");

  _exit (EXIT_SUCCESS);
}

void
Version (void)
{
  _print (STYLE_COMMON, "outpost - v %s\n", VERSION);
  _exit (EXIT_SUCCESS);
}

void
prepare_and_exit ()
{
  struct stat buf;

  unload_all_module ();

  if (logdesc)
    pcap_dump_close (logdesc);

  if (desc)
    pcap_close (desc);

  if (mode_daemon)
  {
    if (!stat (pidfile, &buf))
      unlink (pidfile);

    syslog (SYSLOG_PRIORITY, "shutdowned.");
    closelog ();
  }

  _exit (0);
  return ;
}

void
signal_sigusr1 ()
{
  _print (STYLE_COMMON, "sig usr 1\n");
  return ;
}

void
signal_sigusr2 ()
{
  _print (STYLE_COMMON, "sig usr 2\n");
  return ;
}

void
sig_call (int signal, siginfo_t *siginfo, void *buf)
{
  _print (STYLE_COMMON, "call of signal: %i\n", signal);
  switch (signal)
  {
    case SIGINT: /* interruption, stop. */
      prepare_and_exit ();
    break;
    case SIGUSR1:
      signal_sigusr1 ();
    break;
    case SIGUSR2:
      signal_sigusr2 ();
    break;
    default:
    break;
  }
}

/* ripped from tcpdump's util.c */
char
*preparerule (register char **argv)
{
  register char **p;
  register u_int len = 0;
  char *buf;
  char *src, *dst;
  p = argv;

  if (*p == 0)
    return 0;

  while (*p)
    len += strlen(*p++) + 1;

  buf = (char *)_malloc(len);

  p = argv;
  dst = buf;
  while ((src = *p++) != NULL) 
  {
    while ((*dst++ = *src++) != '\0');
    dst[-1] = ' ';
  }
  dst[-1] = '\0';
  return buf;
}

int
main (int argc, char **argv, char **envp)
{
  int arg;
  int option_index = 0;
  const unsigned char *packet;
  struct pcap_pkthdr pcapstr;

  struct bpf_program fcode;
  bpf_u_int32 localnet, netmask;

  char spid[6];
  int fdpid;

  char errbuf[PCAP_ERRBUF_SIZE];
  int endb = 0;
  int i;

  char *dumpfile = NULL;
  char *readfile = NULL;

  struct sigaction my_sighandler;

  clearconf ();
  
  do
  {
    arg = getopt_long (argc, argv, "cdhi:m:r:s:tvw:IVT", 
      long_options, &option_index);

    switch(arg)
    {
      case 'c':
        mode_test_conf = 1;
      break;
      case 'd':
        mode_daemon = 1;
      break;
      case 'h':
        usage ();
      break;
      case 'i':
        dev = _malloc_strcpy(optarg);
      break;
      case 'm': /* load a module */ 
        load_module_arg (optarg, argc, argv, &option_index);
      break;
      case 'p':
        pidfile = _malloc_strcpy(optarg);
      break;
      case 'r': 
        readfile = _malloc_strcpy(optarg);
      break;
      case 's': 
        snaplen = atoi (optarg);
      break;
      case 't': /* test bpf expression */
        mode_test = 1;
        localnet = 0xffff0000;
      break;
      case 'v': /* verbosity level XXX */

      break;
      case 'w': /* we've to write in a file the dump in pcap format */
                /* pcap format is the same used by tcpdump or snort */
        dumpfile = _malloc_strcpy(optarg);
      break;
      case 'I':
        ignore_conf = 1;
      break;
      case 'T':
        mode_real_test = 1;
      break;
      case 'V': /* Version */
        Version ();
      break;
      default:
      break;
    }
  }
  while (arg != -1);

  if (argv[optind] != NULL)
  {
    cmd = preparerule (&argv[optind]);
    _print (STYLE_COMMON, "BPF rule: %s\n", cmd);
  }

  if (readconf () && !ignore_conf)
  {
    _print (STYLE_ALERT, "fatal: there are errors in file configuration.\n");
    _exit (EXIT_FAILURE);
  }

  initconf ();

  if (mode_test_conf && !ignore_conf)
  {
    _print (STYLE_COMMON, "file configuration seems to be OK.\n");
    _exit (EXIT_SUCCESS);
  }

  if (getuid() && !readfile && !mode_test)
  {
    _print (STYLE_ALERT, "fatal: you've to be root.\n");
    _exit (EXIT_FAILURE);
  }

  /* init signals features */

  my_sighandler.sa_handler = (void *)sig_call;
  my_sighandler.sa_flags = 0;
  sigaction (SIGINT, &my_sighandler, NULL);
  sigaction (SIGUSR1, &my_sighandler, NULL);
  sigaction (SIGUSR2, &my_sighandler, NULL);

  /* init engine */
  if (!dev && !mode_test)
  {
    if (!(dev = pcap_lookupdev (errbuf)))
    {
      _print (STYLE_ALERT, "fatal: no device found by libpcap.\npcap: %s\n", errbuf);
      exit (EXIT_FAILURE);
    }
  }

  if (readfile)
  {
    if (!(desc = pcap_open_offline (readfile, errbuf)))
    {
      _print (STYLE_WARN, "fatal: can't open %s.\npcap: %s\n", readfile, errbuf);
      _free (readfile);
      readfile = NULL;
    }
  }

  /* daemon mode: use of openlog */

  if (mode_daemon)
  {
    openlog (SYSLOG_IDENTITY, LOG_PID | LOG_NDELAY, SYSLOG_FACILITY);
    syslog (SYSLOG_PRIORITY, "started.");
    mode_daemon_e = 1;
  }

  if (mode_test)
    desc = pcap_open_dead (DLT_NULL, snaplen);

  if (!desc)
    desc = pcap_open_live (dev, snaplen, 1, 1000, errbuf);
  if (!desc)
  {
    _print (STYLE_ALERT, "fatal: can't open %s.\npcap: %s\n", dev, errbuf);
    exit (EXIT_FAILURE);
  }

  if (dumpfile)
  {
    logdesc = pcap_dump_open (desc, dumpfile);
    if (!logdesc)
    {
      _print (STYLE_WARN, "pcap_dump_open: %s\n", pcap_geterr (desc));
      _free (dumpfile);
      dumpfile = NULL;
    }
  }

  layer2len = datalinklenbypcap(desc);

  /* bpf engine */

  if (!mode_test)
    if (pcap_lookupnet(dev, &localnet, &netmask, errbuf) < 0)
      _print (STYLE_WARN, "pcap_lookupnet: %s\n", errbuf);

  if (cmd)
  {
    if (pcap_compile(desc, &fcode, cmd, 1, localnet) < 0)
    {
      _print (STYLE_ALERT, "pcap_compile: %s\n", pcap_geterr (desc));
      _exit (EXIT_FAILURE);
    }

    if (mode_test || mode_real_test)
    {
      _print (STYLE_ALERT, "pcap_compile: this rule works very well.\n");
      _exit (EXIT_SUCCESS);
    }
    if (pcap_setfilter (desc, &fcode) < 0)

    {
      _print (STYLE_ALERT, "pcap_setfilter: %s\n", pcap_geterr (desc));
      _exit (EXIT_FAILURE);
    }
  }
  else
    if (mode_test)
    {
      _print(STYLE_ALERT, "pcap_compile: no rule to test ?\n");
      _exit (EXIT_FAILURE);
    }


  /* go in daemon mode */

  if (mode_daemon)
  {
    if (fork () != 0)
      _exit (EXIT_SUCCESS);

    if (setsid () == -1)
      _print (STYLE_WARN, "setsid: %s\n", sys_errlist[errno]);
    else
      for (i = 0; i < NOFILE; i ++)
      {
        /* struct pcap's 1st var is int fd. Hope that doesn't change. */ 
        if (i != pcap_fileno(desc))
          close (i);
      }

    /* pid file */

    fdpid = open (pidfile, O_WRONLY | O_CREAT);
    if (fdpid != -1)
    {
      memset (spid, 0, 6);
      sprintf(spid, "%i", getpid());
      write (fdpid, spid, strlen(spid));
      close (fdpid);
    }
  }

  /* main engine */

  while (!endb)
  {
    packet = pcap_next (desc, &pcapstr); 
    if (packet != NULL)
    {
      if (logdesc)
        pcap_dump ((unsigned char*)logdesc, &pcapstr, packet);

      module_do_action (packet, pcapstr);
    }
    else
    {
      if (readfile)
        endb = 1;
    }
  }

  prepare_and_exit ();

  return (EXIT_SUCCESS);
}
